# NOTE: when we release helios 2.0, this module should be renamed to loggers.py

import copy
import logging
import pathlib
import typing
from datetime import datetime

import matplotlib.pyplot as plt
import numpy.typing as npt
import torch
import torch.utils.tensorboard as tb

from .distributed import get_global_rank
from .utils import get_from_optional


class RootLogger:
    """
    Logger used to log while training to a file.

    The log file will be placed in the logs folder as determined by the given path.

    Args:
        log_file: path to the log file.
        capture_warnings: if true, output of ``warnings.warn`` is captured in the log.
    """

    def __init__(self: "RootLogger", capture_warnings: bool):
        """Create the root logger with stream output as default."""
        logging.captureWarnings(capture_warnings)
        self._logger = logging.getLogger("helios")
        self._rank = get_global_rank()
        self._format_str = "[%(asctime)s] [%(levelname)s]: %(message)s"
        self._log_file: pathlib.Path | None = None

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(self._format_str))
        stream_handler.setLevel(logging.WARNING)
        self._logger.addHandler(stream_handler)
        self._logger.propagate = False

        if self._rank != 0:
            self._logger.setLevel(logging.ERROR)
        else:
            self._logger.setLevel(logging.INFO)

    def setup(self, log_file: pathlib.Path | None = None) -> None:
        """
        Finish configuring the root logger.

        In particular, this function will create the file logger provided that the input
        path is not ``None``. If the path points to a file that already exists, then the
        logger will automatically append to the file, otherwise a new file will be
        created.

        Args:
            log_file: the (optional) path to the log file.
        """
        if log_file is None:
            return

        mode = "a" if log_file.exists() else "w"
        file_handler = logging.FileHandler(str(log_file), mode=mode, encoding="utf-8")
        file_handler.setFormatter(logging.Formatter(self._format_str))
        file_handler.setLevel(logging.INFO)
        self._logger.addHandler(file_handler)
        self._log_file = log_file

    @property
    def logger(self) -> logging.Logger:
        """Return the logger instance."""
        return self._logger

    @property
    def log_file(self) -> pathlib.Path | None:
        """
        Return the path to the current log file.

        If the path for the log file was originally ``None``, this will return ``None``
        as well.
        """
        return self._log_file

    def info(self, msg: str, **kwargs: typing.Any) -> None:
        """
        Log using the ``INFO`` tag.

        Only available for the main process in distributed training.

        Args:
            msg: the message to log
            kwargs: keyword arguments to logging.info
        """
        if self._rank != 0:
            return

        self._logger.info(msg, **kwargs)

    def warning(self, msg: str, **kwargs: typing.Any) -> None:
        """
        Log using the ``WARNING`` tag.

        Args:
            msg: the message to log
            kwargs: keyword arguments to logging.warning
        """
        self._logger.warning(msg)

    def error(self, msg: str, **kwargs: typing.Any) -> None:
        """
        Log using the ``ERROR`` tag.

        Args:
            msg (str): the message to log
            kwargs (Any): keyword arguments to logging.error
        """
        self._logger.error(msg)

    def exception(self, msg: str, **kwargs: typing.Any) -> None:
        """
        Log an exception.

        Args:
            msg: the message to log
            kwargs: keyword arguments to ``logging.exception``
        """
        self._logger.exception(msg, **kwargs)

    def flush(self) -> None:
        """Flush the log."""
        for handler in self._logger.handlers:
            handler.flush()

    def close(self) -> None:
        """Close the log."""
        for handler in self._logger.handlers:
            handler.close()


class TensorboardWriter:
    """
    Wrapper for the ``Tensorboard`` writer.

    Args:
        run_path: path to the directory for the run.
    """

    def __init__(self: "TensorboardWriter"):
        """Create the ``Tensorboard`` writer."""
        self._rank = get_global_rank()
        self._writer: tb.SummaryWriter | None = None
        self._run_path: pathlib.Path | None = None

    def setup(self, run_path: pathlib.Path) -> None:
        """
        Finish configuring the ``TensorboardWriter``.

        In particular, this function will create the writer instance and assign it to the
        given path. If the path already exists, Tensorboard will automatically append to
        the previous run. In distributed training, the writer will only be created on rank
        0.

        Args:
            run_path:: the path to the run folder.
        """
        if self._rank == 0:
            self._writer = tb.SummaryWriter(log_dir=str(run_path), flush_secs=20)
            self._run_path = run_path

    @property
    def run_path(self) -> pathlib.Path:
        """Return the path to the current run folder."""
        return get_from_optional(self._run_path)

    def add_scalar(
        self,
        tag: str,
        scalar_value: float | str | torch.Tensor,
        global_step: int | None = None,
    ) -> None:
        """
        Add scalar data to the log.

        Args:
            tag: name of the scalar to plot.
            scalar_value: the scalar to plot.
            global_step: the (optional) step for the given scalar.
        """
        if self._writer is None:
            return

        self._writer.add_scalar(tag, scalar_value, global_step, new_style=True)

    def add_scalars(
        self,
        main_tag: str,
        tag_scalar_dict: dict[str, float | torch.Tensor],
        global_step: int | None,
    ) -> None:
        """
        Add multiple scalars to the log.

        Args:
            main_tag: the parent name for the tags
            tag_scalar_dict: key-value pair storing tag and corresponding values.
            global_step: (optional) global step value to record.
        """
        if self._writer is None:
            return

        self._writer.add_scalars(main_tag, tag_scalar_dict, global_step)

    def add_image(
        self,
        tag: str,
        img_tensor: torch.Tensor | npt.NDArray,
        global_step: int | None = None,
        dataformats: str = "CHW",
    ) -> None:
        """
        Add image data to log.

        Args:
            tag: data identifier.
            img_tensor: image data.
            global_step: (optional) global step value to record
            dataformats: image data format specification in the form CHW, HWC, HW, WH,
                etc.
        """
        if self._writer is None:
            return

        self._writer.add_image(tag, img_tensor, global_step, dataformats=dataformats)

    def add_images(
        self,
        tag: str,
        img_tensor: torch.Tensor | npt.NDArray,
        global_step: int | None = None,
        dataformats: str = "NCHW",
    ) -> None:
        """
        Add batched image data to log.

        Args:
            tag: data identifier.
            img_tensor: image data.
            global_step: (optional) global step value to record.
            dataformats: image data format specification in the form NCHW, NHWC, CHW,
                HWC, HW, WH, etc.
        """
        if self._writer is None:
            return

        self._writer.add_images(tag, img_tensor, global_step, dataformats=dataformats)

    def add_figure(
        self,
        tag: str,
        figure: plt.Figure | list[plt.Figure],
        global_step: int | None = None,
        close: bool = True,
    ) -> None:
        """
        Render matplotlib figure into an image and add it to a summary.

        Args:
            tag: data identifier.
            figure: figure or a list of figures.
            global_step: (optional) global step value to record.
            close: flag to automatically close the figure.
        """
        if self._writer is None:
            return

        self._writer.add_figure(tag, figure, global_step, close)

    def add_text(self, tag: str, text_string: str, global_step: int | None = None):
        """
        Add text data to log.

        Args:
            tag: data identifier.
            text_string: string to save.
            global_step: (optional) global step value to record.
        """
        if self._writer is None:
            return

        self._writer.add_text(tag, text_string, global_step)

    def add_graph(
        self,
        model: torch.nn.Module,
        input_to_model: torch.Tensor | list[torch.Tensor] | None = None,
        verbose: bool = False,
        use_strict_trace: bool = True,
    ) -> None:
        """
        Add graph data to log.

        PyTorch currently supports CPU tracing only, so the model and its input(s) will be
        automatically moved to the CPU prior to logging. Note that the model will be
        copied prior to moving to the CPU to ensure the input model is not affected.

        Args:
            model: model to draw.
            input_to_mode: a variable or tuple of variables to be fed.
            verbose: whether to print graph structure in console.
            use_strict_trace: whether to pass keyword argument strict to torch.jit.trace.
        """
        if self._writer is None:
            return

        # Transfer graph + input to the CPU for tracing.
        model_cpu = copy.deepcopy(model).cpu()

        input_cpu: torch.Tensor | list[torch.Tensor] | None
        if isinstance(input_to_model, torch.Tensor):
            input_cpu = input_to_model.cpu()
        elif isinstance(input_to_model, list):
            input_cpu = []
            for x in input_to_model:
                input_cpu.append(x.cpu())
        else:
            input_cpu = None

        self._writer.add_graph(model_cpu, input_cpu, verbose, use_strict_trace)

    def add_pr_curve(
        self,
        tag: str,
        labels: torch.Tensor | npt.NDArray,
        predictions: torch.Tensor | npt.NDArray,
        global_step: int | None = None,
        num_thresholds: int = 127,
    ) -> None:
        """
        Add a precision recall curve to the log.

        Plotting a precision-recall curve lets you understand your model's performance
        under different threshold settings. With this function, you provide the ground
        truth labelling (T/F) and prediction confidence (usually the output of your model)
        for each target. The TensorBoard UI will let you choose the threshold
        interactively.

        Args:
            tag: data identifier.
            labels: ground truth data. Binary label for each element.
            predictions: the probability that an element be classified as true. Value
                should be in [0, 1].
            global_step: (optional) global step value to record.
            num_thresholds: number of thresholds used to draw the curve. Defaults to 127.
        """
        if self._writer is None:
            return

        self._writer.add_pr_curve(tag, labels, predictions, global_step, num_thresholds)

    def add_hparams(
        self,
        hparam_dict: dict[str, typing.Any],
        metric_dict: dict[str, typing.Any],
        hparam_domain_discrete: dict[str, list[typing.Any]] | None = None,
        run_name: str | None = None,
        global_step: int | None = None,
    ) -> None:
        """
        Add a set of hyper-parameters to be compared in the log.

        Args:
            hparam_dict: each key-value pair in the dictionary is the name of the
                hyper-parameter and it's corresponding value. The type of the value can
                be one of ``bool``, ``string``, ``float``, ``int``, or ``None``.
            metric_dict: each key-value pair in the dictionary is the name of the metric
                and it's corresponding value. Note that the key used here should be
                unique in the Tensorboard record.
            hparam_domain_discrete:: a dictionary that contains names of hyper-parameters
                and all the discrete values they can hold.
            run_name: name of the run, to be included as part of the logdir.
            global_step: (optional) global step value to record.
        """
        if self._writer is None:
            return

        if run_name is None:
            run_name = "."

        self._writer.add_hparams(
            hparam_dict, metric_dict, hparam_domain_discrete, run_name, global_step
        )

    def flush(self) -> None:
        """Flush any cached values to Tensorboard."""
        if self._writer is None:
            return

        self._writer.flush()

    def close(self) -> None:
        """Close the Tensorboard writer."""
        if self._writer is None:
            return

        self._writer.close()


_ACTIVE_LOGGERS: dict[str, RootLogger | TensorboardWriter] = {}


def get_default_log_name(run_name: str) -> str:
    """
    Generate the default name used for log files or Tensorboard runs.

    Args:
        run_name: the name of the run.

    Returns:
        The string with the default name.
    """
    current_time = datetime.now().strftime("%b%d_%H-%M-%S")

    return run_name + f"_{current_time}"


def create_default_loggers(
    enable_tensorboard: bool = True, capture_warnings: bool = True
) -> None:
    """
    Construct the ``RootLogger`` and ``TensorboardWriter`` instances.

    In distributed training, this function should be called AFTER the processes have been
    created to ensure each process gets a copy of the loggers.

    Args:
        enable_tensorboard: (optional) enables the tensorboard writer. Defaults to true.
    """
    # Root logger is always created.
    if "root" not in _ACTIVE_LOGGERS:
        _ACTIVE_LOGGERS["root"] = RootLogger(capture_warnings=capture_warnings)

    if enable_tensorboard and "tensorboard" not in _ACTIVE_LOGGERS:
        _ACTIVE_LOGGERS["tensorboard"] = TensorboardWriter()


def restore_default_loggers(
    log_path: pathlib.Path | None = None, run_path: pathlib.Path | None = None
) -> None:
    """
    Restore the default loggers from a previous run.

    This function should be called whenever the loggers need to continue logging to the
    same file/folder as a previous run.

    Args:
        log_path: (optional) path to the log file.
        run_path: (optional) path to the Tensorboard run folder.

    Raises:
        RuntimeError: if the root logger hasn't been created.
    """
    if "root" not in _ACTIVE_LOGGERS:
        raise RuntimeError(
            "error: root logger hasn't been created. Did you forget to call "
            "create_default_loggers?"
        )

    if log_path is not None:
        _ACTIVE_LOGGERS["root"].setup(log_path)

    if "tensorboard" in _ACTIVE_LOGGERS and run_path is not None:
        _ACTIVE_LOGGERS["tensorboard"].setup(run_path)


def setup_default_loggers(
    run_name: str,
    log_root: pathlib.Path | None = None,
    runs_root: pathlib.Path | None = None,
) -> None:
    """
    Call the setup functions on the default loggers.

    This function should be called when the loggers don't need to continue from a previous
    run. If you need that, call restore_default_loggers instead.

    Args:
        run_name: the name of the current run.
        log_root: (optional) path to the logs folder.
        runs_root: (optional) path to the Tensorboard runs folder.
    """
    base_name = get_default_log_name(run_name)
    run_path = None if runs_root is None else runs_root / base_name
    log_path = None if log_root is None else log_root / f"{base_name}.log"
    restore_default_loggers(log_path, run_path)


def flush_default_loggers() -> None:
    """Flushes the default loggers."""
    for _, logger in _ACTIVE_LOGGERS.items():
        logger.flush()


def close_default_loggers() -> None:
    """Close the default loggers and clears them."""
    for _, logger in _ACTIVE_LOGGERS.items():
        logger.close()

    _ACTIVE_LOGGERS.clear()


def get_root_logger() -> RootLogger:
    """
    Get the root logger instance.

    Return:
        The root logger.
    """
    if "root" not in _ACTIVE_LOGGERS:
        raise KeyError(
            "error: root logger has not been created. Did you forget to call "
            "create_default_loggers?"
        )
    return typing.cast(RootLogger, _ACTIVE_LOGGERS["root"])


def is_root_logger_active() -> bool:
    """Check if the root logger has been created."""
    return "root" in _ACTIVE_LOGGERS


def get_tensorboard_writer() -> TensorboardWriter | None:
    """
    Return the Tensorboard writter.

    If Tensorboard is disabled, this function will return ``None``

    Return:
        The Tensorboard logger or ``None`` if it doesn't exist.
    """
    if "tensorboard" not in _ACTIVE_LOGGERS:
        return None
    return typing.cast(TensorboardWriter, _ACTIVE_LOGGERS["tensorboard"])



import typing

import torch

import helios.model as hlm
import helios.plugins as hlp
import helios.trainer as hlt


class ExceptionPlugin(hlp.Plugin):
    def __init__(self, exc_type: type[Exception] | list[type[Exception]]):
        super().__init__(f"exception_{str(exc_type)}")
        self._exc_type = exc_type

    def setup(self) -> None:
        pass

    def configure_trainer(self, t: hlt.Trainer) -> None:
        self._append_train_exceptions(self._exc_type, t)
        self._append_test_exceptions(self._exc_type, t)


class PluginModel(hlm.Model):
    def __init__(self, save_name: str):
        super().__init__(save_name)

    def setup(self, fast_init: bool = False) -> None:
        pass


class TestPlugins:
    def test_registry(self, check_registry) -> None:
        check_registry(hlp.PLUGIN_REGISTRY, ["CUDAPlugin", "OptunaPlugin"])

    def test_create(self, check_create_function) -> None:
        check_create_function(hlp.PLUGIN_REGISTRY, hlp.create_plugin)

    def test_append_exceptions(self) -> None:
        t = hlt.Trainer()
        plugin = ExceptionPlugin(RuntimeError)
        exc_list: list[type[Exception]] = [RuntimeError]

        plugin.configure_trainer(t)
        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list

        exc_list.extend([ValueError, TypeError])
        plugin = ExceptionPlugin([ValueError, TypeError])
        plugin.configure_trainer(t)

        assert t.train_exceptions == exc_list
        assert t.test_exceptions == exc_list


class TestCUDAPlugin:
    def check_batch_device(self, x: typing.Any, device: torch.device) -> None:
        if isinstance(x, torch.Tensor):
            assert x.device == device
        elif isinstance(x, dict):
            assert all(val.device == device for _, val in x.items())
        else:
            assert all(elem.device == device for elem in x)

    def check_batch_processing(
        self, plugin: hlp.CUDAPlugin, x: typing.Any, device: torch.device
    ) -> None:
        ret = plugin.process_training_batch(x, hlt.TrainingState())
        self.check_batch_device(ret, device)

        ret = plugin.process_validation_batch(x, 0)
        self.check_batch_device(ret, device)

        ret = plugin.process_testing_batch(x, 0)
        self.check_batch_device(ret, device)

    def test_plugin_id(self) -> None:
        assert hasattr(hlp.CUDAPlugin, "plugin_id")
        assert hlp.CUDAPlugin.plugin_id == "cuda"

    def test_batch_processing(self) -> None:
        if not torch.cuda.is_available():
            return

        def create_tensor() -> torch.Tensor:
            return torch.randn((1, 3, 32, 32)).to("cpu")

        device = torch.device("cuda:0")

        plugin = hlp.CUDAPlugin()
        plugin.is_distributed = False
        plugin.map_loc = {"cuda:0": "cuda:0"}
        plugin.device = device

        assert plugin.unique_overrides.training_batch
        assert plugin.unique_overrides.validation_batch
        assert plugin.unique_overrides.testing_batch

        self.check_batch_processing(plugin, create_tensor(), device)
        self.check_batch_processing(plugin, [create_tensor(), create_tensor()], device)
        self.check_batch_processing(plugin, (create_tensor(), create_tensor()), device)
        self.check_batch_processing(
            plugin, {"a": create_tensor(), "b": create_tensor()}, device
        )

    def test_configure(self) -> None:
        if not torch.cuda.is_available():
            return

        trainer = hlt.Trainer()
        plugin = hlp.CUDAPlugin()
        plugin.configure_trainer(trainer)

        assert "cuda" in trainer.plugins
        assert trainer.plugins["cuda"] == plugin



"""
Neural Network package for Helios.

This contains the classes, utilities, and registries related to neural networks.
"""

from . import swa_utils
from .utils import NETWORK_REGISTRY, create_network, default_init_weights

__all__ = [
    "swa_utils",
    "NETWORK_REGISTRY",
    "create_network",
    "default_init_weights",
]



from __future__ import annotations

import abc
import typing

import torch

from helios import core
from helios.core import distributed as dist

if typing.TYPE_CHECKING:
    from ..trainer import Trainer, TrainingState


# Tell Ruff to ignore the empty-method-without-abstract-decorator check, since a lot of
# functions in the Model base class will be empty by default.
# ruff: noqa: B027


class Model(abc.ABC):
    """
    Base class that groups together the functionality needed to train networks.

    The use of this class is to standardize the way networks are created and trained. This
    allows the training code to be shared across multiple networks, reducing code
    duplication.
    The functions provided by the :py:class:`~helios.model.model.Model` class can be
    overridden to satisfy the individual needs of each of the network(s) that need to be
    trained.

    Example:
        Suppose the body of the training loop looks something like this:

        .. code-block:: python

            dataloader = ... # The dataloader for our dataset.
            net = ... # The network we wish to train.
            optimzer = ... # The optimizer
            criterion = ... # The loss function
            for batch in dataloder:
                inputs, labels = batch

                optimizer.zero_grad()
                outs = net(inputs)
                loss = criterion(outs, labels)
                loss.backward()
                optimizer.step()

        Then the code would be placed into a Model as follows:

        .. code-block:: python

            import helios.model as hlm
            import helios.trainer as hlt
            class MyModel(hlm.Model):
                def setup(self, fast_init: bool = False) -> None:
                    self._net = ...
                    self._optimizer = ...
                    self._criterion = ...

                def train_step(self, batch, state: hlt.TrainingState) -> None:
                    inputs, labels = batch

                    optimizer.zero_grad()
                    outs = net(inputs)
                    loss = criterion(outs, labels)
                    loss.backward()
                    optimizer.step()

    The example shown here is the most basic version of the training code and can be
    expanded in various ways by overriding the different available functions.

    Args:
        save_name: the name that will be used to identify the model when checkpoints are
            saved.
    """

    def __init__(self, save_name: str):
        """Create the model."""
        self._save_name: str = save_name
        self._trainer: Trainer | None

        self._is_distributed: bool = False
        self._map_loc: str | dict[str, str] = ""
        self._device: torch.device | None = None
        self._rank: int = 0

        self._loss_items: dict[str, torch.Tensor] = {}
        self._running_losses: dict[str, float] = {}
        self._val_scores: dict[str, typing.Any] = {}
        self._test_scores: dict[str, typing.Any] = {}

    @property
    def save_name(self) -> str:
        """The name of the model used for saving checkpoints and final networks."""
        return self._save_name

    @property
    def is_distributed(self) -> bool:
        """Flag controlling whether distributed training is being used or not."""
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, val: bool) -> None:
        self._is_distributed = val

    @property
    def map_loc(self) -> str | dict[str, str]:
        """The location to map loaded weights from a checkpoint or pre-trained file."""
        return self._map_loc

    @map_loc.setter
    def map_loc(self, loc: str | dict[str, str]) -> None:
        self._map_loc = loc

    @property
    def device(self) -> torch.device:
        """The device on which the tensors of the model are mapped to."""
        return core.get_from_optional(self._device)

    @device.setter
    def device(self, dev: torch.device) -> None:
        self._device = dev

    @property
    def rank(self) -> int:
        """The local rank (device id) that the model is running on."""
        return self._rank

    @rank.setter
    def rank(self, r: int) -> None:
        self._rank = r

    @property
    def trainer(self) -> Trainer:
        """Reference to the trainer."""
        return core.get_from_optional(self._trainer)

    @trainer.setter
    def trainer(self, t) -> None:
        self._trainer = t

    @property
    def metrics(self) -> dict[str, typing.Any]:
        """The metric table for the model."""
        return {}

    @abc.abstractmethod
    def setup(self, fast_init: bool = False) -> None:
        """
        Initialize all the state necessary for training.

        Use this function to load all the networks, schedulers, optimizers, losses, etc.
        that you require for training. This will be called before training starts and
        after the distributed processes have been launched (if applicable).

        The ``fast_init`` flag is used to indicate that the model should **not** load any
        training state. This can be used for testing or for other purposes.

        Args:
            fast_init: if True, only networks are loaded.
        """

    def load_for_testing(self) -> None:
        """
        Load the network(s) used for testing.

        When testing, the trainer will try to load the last checkpoint saved during
        training. If it is unable to find one, it will call this function to ensure the
        model loads the state of the network(s) for testing. If you wish to load your
        network using separate logic, use this function.
        """

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        """
        Load the model state from the given state dictionary.

        Use this function to restore any training state from a checkpoint. Note that any
        weights will have been automatically mapped to the correct device.

        The ``fast_init`` flag is used to indicate that the model should **not** load any
        training state. This can be used for testing or for other purposes. As such, you
        should only load the state of your network(s) and nothing else.

        Args:
            state_dict: the state dictionary to load from.
            fast_init: if True, only networks need to be loaded.
        """

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state dictionary of the model.

        Use this function to save any state that you require for checkpoints.

        Returns:
            The state dictionary of the model.
        """
        return {}

    def trained_state_dict(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> dict[str, typing.Any]:
        """
        Get the state dictionary for the trained model.

        Use this function to save the state required for the final trained model. The
        returned dictionary should contain only the necessary information to re-create the
        network(s) along with any additional data you require.

        Args:
            args: positional arguments.
            kwargs: keyword arguments.

        Returns:
            The state dictionary without any training data.
        """
        return {}

    def types_for_safe_load(self) -> list[typing.Callable | tuple[typing.Callable, str]]:
        """
        Get additional types to safely load checkpoints.

        These types are appended to the global list of types from Helios and are sent to
        ``torch.serialization.add_safe_globals`` so checkpoints can be safely loaded.

        Returns:
            The list of safe types to load.
        """
        return []

    def append_metadata_to_chkpt_name(self, chkpt_name: str) -> str:
        """
        Append additional data to the checkpoint filename.

        Use this function to append the value of the loss function(s), validation
        metric(s), or any extra metadata you wish to add to the name of the checkpoint.

        .. note::
            The epoch and iteration numbers, alongside the file extension, are added
            automatically.

        Args:
            chkpt_name: the name of the checkpoint filename (without extension).

        Returns:
            The name with any additional metadata.
        """
        return chkpt_name

    def append_to_banner(self, banner: str) -> str:
        """
        Append additional information to the main banner printed on start-up.

        Use this function to add any extra information you wish to the main banner shown
        at start-up. Note that if ``print_banner`` is set to false in the
        :py:class:`~helios.trainer.Trainer`, then this message is not shown.

        Returns:
            The banner string with any additional information.
        """
        return banner

    def train(self) -> None:
        """Switch the model to training mode."""

    def on_training_start(self) -> None:
        """
        Perform any necessary actions when training starts.

        You may use this function to log the network architecture, hyper-params, etc.
        """

    def on_training_epoch_start(self, current_epoch: int) -> None:
        """
        Perform any necessary actions when a training epoch is started.

        This function is called whenever a new training epoch begins, at the top of the
        training loop. You may use this function to set any necessary training state.

        Args:
            current_epoch: the epoch number that has just started.
        """

    def on_training_batch_start(self, state: TrainingState) -> None:
        """
        Perform any actions when a training batch is started.

        This function is called before :py:meth:`train_step` is called. By default, it
        will clear out the loss table, but you may also use it to do any additional tasks
        prior to the training step itself.

        Args:
            state: the current training state.
        """
        self._loss_items.clear()

    def train_step(self, batch: typing.Any, state: TrainingState) -> None:
        """
        Perform a single training step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform the forward and backward passes for your
        network(s). If you use schedulers, they should be updated here as well. Note that
        you do not have to clear the losses or gather them. This will be handled
        automatically for you.

        .. warning::
            The contents of the batch **are not** moved to any devices prior to this call.
            It is your responsibility to move them (if necessary).

        Args:
            batch: the batch data returned from the dataset.
            state: the current training state.
        """

    def on_training_batch_end(
        self, state: TrainingState, should_log: bool = False
    ) -> None:
        """
        Perform any actions when a training batch ends.

        This function is called after :py:meth:`~helios.model.model.Model.train_step` is
        called. By default, it will gather all the losses stored in ``self._loss_items``
        (if using distributed training) and will update the running losses using those
        values. You may also use this function to log your losses or perform any
        additional tasks after the training step.

        Args:
            state: the current training state.
            should_log: if true, then logging should be performed. Defaults to false.
        """
        if self._is_distributed:
            for _, loss in self._loss_items.items():
                dist.all_reduce_tensors(loss)

        for key, loss in self._loss_items.items():
            if loss is not None:
                if key not in self._running_losses:
                    self._running_losses[key] = loss.item()
                else:
                    self._running_losses[key] += loss.item()

    def on_training_epoch_end(self, current_epoch: int) -> None:
        """
        Perform any necessary actions when a training epoch ends.

        This function is called at the bottom of the epoch loop. You may use this
        function to perform any training operations you require.

        Args:
            current_epoch: the epoch number that has just ended.
        """

    def on_training_end(self) -> None:
        """
        Perform any necessary actions when training ends.

        You may use this function to update any weight averaging networks, or any other
        tasks that should only happen at the end of training.
        """

    def eval(self) -> None:
        """Switch the model to evaluation mode."""

    def on_validation_start(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation starts.

        By default, this will clear out the table of validation values, but you may use it
        for any other tasks that should happen when validation begins.

        Args:
            validation_cycle: the validation cycle number.
        """
        self._val_scores.clear()

    def on_validation_batch_start(self, step: int) -> None:
        """
        Perform any actions when a validation batch is started.

        This function is called before :py:meth:`valid_step` is called. No steps are
        performed by default.

        Args:
            step: the current validation batch.
        """

    def valid_step(self, batch: typing.Any, step: int) -> None:
        """
        Perform a single validation step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform any steps necessary to compute the validation
        metric(s) for your network(s).

        Args:
            batch: the batch data returned from the dataset.
            step: the current validation batch.
        """

    def on_validation_batch_end(self, step: int) -> None:
        """
        Perform any actions when a validation batch ends.

        This function is called after :py:meth:`valid_step` is called. No steps are
        performed by default.

        Args:
            step: the current validation batch.
        """

    def on_validation_end(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation ends.

        By default, this function will clear out the running loss table, but you may use
        this function to compute any final validation metrics as well as log them.

        Args:
            validation_cycle: the validation cycle number.
        """
        self._running_losses.clear()

    def have_metrics_improved(self) -> bool:
        """
        Determine whether the current validation results are an improvement or not.

        This is used when early stopping is enabled in the trainer to determine whether
        the stop cycle count should increase or not. This is called immediately after the
        validation cycle finishes.

        Returns:
            False if no improvements were seen in the last validation cycle.
        """
        return True

    def should_training_stop(self) -> bool:
        """
        Determine whether training should stop or continue.

        This is used in the event that training should stop when:

        * A validation metric crosses a certain threshold,
        * A loss value becomes invalid,
        * Any other circumstance under which training should stop immediately.

        This function is called by the :py:class:`~helios.trainer.Trainer` at the
        following times:

        * After a full training step has finished. That is, after
          :py:meth:`on_training_batch_start`, :py:meth:`train_step`, and
          :py:meth:`on_training_batch_end` have been called.
        * After a validation cycle has finished. That is, after *all* of the validation
          callbacks have been called.
        * After each epoch has concluded (if training by epoch). Note that this happens
          *after* :py:meth:`on_training_epoch_end`.

        Returns:
            False if training should continue, true otherwise.
        """
        return False

    def on_testing_start(self) -> None:
        """
        Perform any necessary actions when testing starts.

        By default, this will clear out the table of testing values, but you may use it
        for any other tasks that should happen when testing begins.
        """
        self._test_scores.clear()

    def on_testing_batch_start(self, step: int) -> None:
        """
        Perform any actions when a testing batch is started.

        This function is called before :py:meth:`~helios.model.model.Model.test_step` is
        called. No steps are performed by default.

        Args:
            step: the current testing batch.
        """

    def test_step(self, batch: typing.Any, step: int) -> None:
        """
        Perform a single testing step.

        The input is the returned value from the datasets you supplied to the trainer. In
        this function, you should perform any steps necessary to compute the testing
        metric(s) for your network(s).

        Args:
            batch: the batch data returned from the dataset.
            step: the current validation batch.
        """

    def on_testing_batch_end(self, step: int) -> None:
        """
        Perform any actions when a testing batch ends.

        This function is called after :py:meth:`~helios.model.model.Model.test_step` is
        called. No steps are performed by default.

        Args:
            step: the current testing batch.
        """

    def on_testing_end(self) -> None:
        """
        Perform any necessary actions when testing ends.

        You may use this function to compute any final testing metrics as well as log
        them.
        """



"""
Optimizer package for Helios.

This contains the classes, utilities, and registries related to optimizers.
"""

from .utils import OPTIMIZER_REGISTRY, create_optimizer

__all__ = ["OPTIMIZER_REGISTRY", "create_optimizer"]



import pathlib
import typing

import cv2
import numpy as np
import numpy.typing as npt
import PIL
import torch
from torch.utils import data as tud

from helios import core, data
from helios.core import rng
from helios.data import functional as hldf
from helios.data import samplers as hlds
from helios.data import transforms as hldt

DATASET_SIZE: int = 16


@data.DATASET_REGISTRY.register
class RandomDataset(tud.Dataset):
    def __getitem__(self, index):
        gen = rng.get_default_numpy_rng().generator
        return gen.integers(0, 1000, 3)

    def __len__(self):
        return DATASET_SIZE


@data.DATASET_REGISTRY.register
class SequentialDataset(tud.Dataset):
    def __init__(self):
        super().__init__()

        self._samples = [
            torch.tensor([359, 639, 636]),
            torch.tensor([963, 738, 996]),
            torch.tensor([766, 97, 292]),
            torch.tensor([500, 324, 687]),
            torch.tensor([15, 841, 180]),
            torch.tensor([553, 266, 928]),
            torch.tensor([269, 806, 472]),
            torch.tensor([694, 632, 678]),
            torch.tensor([92, 228, 424]),
            torch.tensor([421, 568, 705]),
            torch.tensor([686, 769, 903]),
            torch.tensor([440, 610, 126]),
            torch.tensor([809, 833, 629]),
            torch.tensor([340, 810, 512]),
            torch.tensor([23, 322, 577]),
            torch.tensor([757, 651, 429]),
        ]

    def __getitem__(self, index):
        return self._samples[index]

    def __len__(self):
        return DATASET_SIZE


class SampleDataModule(data.DataModule):
    def setup(self) -> None:
        params = data.DataLoaderParams(
            batch_size=2, num_workers=2, random_seed=rng.get_default_seed()
        )

        params.shuffle = True
        self._train_dataset = self._create_dataset(RandomDataset(), params)

        params.num_workers = 2
        self._valid_dataset = self._create_dataset(SequentialDataset(), params)

        params.shuffle = False
        self._test_dataset = self._create_dataset(SequentialDataset(), params)


def dummy_collate_fn(a: int) -> int:
    return a + 1


@data.COLLATE_FN_REGISTRY.register
class SampleCollateFn:
    def __init__(self, seed: int) -> None:
        self.seed = seed

    def __call__(self) -> int:
        return self.seed + 1


class TestTransforms:
    def check_type(self, x, exp_type: type) -> None:
        assert isinstance(x, exp_type)

        if isinstance(exp_type, list | tuple):
            for elem in x:  # type: ignore[attr-defined]
                assert isinstance(elem, torch.Tensor)

    def check_to_tensor(self, img, exp_type: type, exp_shape: tuple[int, ...]) -> None:
        to_tensor = hldt.create_transform("ToImageTensor")
        as_tens: torch.Tensor | list[torch.Tensor] | tuple[torch.Tensor, ...] = to_tensor(
            img
        )

        self.check_type(as_tens, exp_type)

        if isinstance(as_tens, list | tuple):
            for t in as_tens:
                assert t.shape == exp_shape
        else:
            assert as_tens.shape == exp_shape

    def generate_sample(
        self,
        size: tuple[int, ...],
        num_elems: int,
        as_tuple: bool = False,
        as_pil: bool = False,
    ):
        gen = rng.get_default_numpy_rng().generator

        samples: list[npt.NDArray | PIL.Image.Image] = []
        for _ in range(num_elems):
            base = np.uint8(gen.random(size=size)) * 255
            if as_pil:
                samples.append(PIL.Image.fromarray(base))
            else:
                samples.append(base)  # type: ignore[arg-type]

        if as_tuple:
            return tuple(samples)
        return samples

    def test_to_tensor(self) -> None:
        assert "ToImageTensor" in hldt.TRANSFORM_REGISTRY

        rng.seed_rngs()
        size = (32, 32, 3)
        exp_size = (3, 32, 32)

        # Single image (numpy)
        self.check_to_tensor(self.generate_sample(size, 1)[0], torch.Tensor, exp_size)

        # Multi-image list (numpy)
        self.check_to_tensor(self.generate_sample(size, 2), list, exp_size)

        # Multi-image tuple (numpy)
        self.check_to_tensor(
            self.generate_sample(size, 2, as_tuple=True),
            tuple,
            exp_size,
        )

        # Single image (PIL)
        self.check_to_tensor(
            self.generate_sample(size, 1, as_pil=True)[0], torch.Tensor, exp_size
        )

        # Multi-image list (PIL)
        self.check_to_tensor(self.generate_sample(size, 2, as_pil=True), list, exp_size)

        # Multi-image tuple (PIL)
        self.check_to_tensor(
            self.generate_sample(size, 2, as_tuple=True, as_pil=True),
            tuple,
            exp_size,
        )


class TestFunctional:
    def check_image(self, gt: npt.NDArray, out_path: pathlib.Path, **kwargs) -> None:
        # Ensure that we convert to "BGR" so the output comes out the way we expect it to
        # come out.
        if kwargs.get("as_rgb", True):
            if len(gt.shape) == 2:
                tmp = gt
            elif len(gt.shape) == 3 and gt.shape[2] == 3:
                tmp = cv2.cvtColor(gt, cv2.COLOR_RGB2BGR)
            elif len(gt.shape) == 3 and gt.shape[2] == 4:
                tmp = cv2.cvtColor(gt, cv2.COLOR_RGBA2BGRA)
        else:
            tmp = gt

        cv2.imwrite(str(out_path), tmp)

        ret = hldf.load_image(out_path, **kwargs)
        assert isinstance(ret, np.ndarray)
        assert np.array_equal(gt, ret)

    def test_load_image(self, tmp_path: pathlib.Path):
        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        out_path = tmp_path / "tmp.png"
        max_8 = 255
        max_16 = 65535

        # 8-bit grayscale.
        self.check_image(
            (gen.random(size=(32, 32)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_GRAYSCALE,
        )

        # 16-bit grayscale
        self.check_image(
            (gen.random(size=(32, 32)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 8-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 3)) * max_8).astype(np.uint8),
            out_path,
        )

        # 16-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 3)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 8-bit RGBA
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # 16-bit RGB
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_16).astype(np.uint16),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
        )

        # Load image as is.
        self.check_image(
            (gen.random(size=(32, 32, 4)) * max_8).astype(np.uint8),
            out_path,
            flags=cv2.IMREAD_UNCHANGED,
            as_rgb=False,
        )

    def test_load_image_pil(self, tmp_path: pathlib.Path):
        rng.seed_rngs()
        gen = rng.get_default_numpy_rng().generator
        img = (gen.random(size=(32, 32, 3)) * 255).astype(np.uint8)
        as_pil = PIL.Image.fromarray(img)

        out_path = tmp_path / "tmp.png"
        cv2.imwrite(str(out_path), cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

        ret = hldf.load_image_pil(out_path)
        assert isinstance(ret, np.ndarray)
        assert np.array_equal(img, ret)

        ret = hldf.load_image_pil(out_path, as_numpy=False)
        assert isinstance(ret, PIL.Image.Image)
        diff = PIL.ImageChops.difference(as_pil, ret)  # type: ignore[attr-defined]
        assert diff.getbbox() is None


class TestTensorToNumpy:
    def get_default_tensor_args(self) -> dict[str, typing.Any]:
        return {
            "squeeze": False,
            "clamp": None,
            "transpose": False,
            "dtype": None,
            "as_uint8": False,
        }

    def test_clamp(self) -> None:
        # Base case: no clamping.
        args = self.get_default_tensor_args()
        x = torch.FloatTensor([-1.0, 1.0])
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert np.max(ret) == 1.0
        assert np.min(ret) == -1.0

        # Clamp to 0.5
        args["clamp"] = (0, 0.5)
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert np.max(ret) == 0.5
        assert np.min(ret) == 0.0

    def test_squeeze(self) -> None:
        # Base case: no squeezing
        args = self.get_default_tensor_args()
        x = torch.randn((1, 32, 1, 32))
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (1, 32, 1, 32)

        # Squeeze
        args["squeeze"] = True
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (32, 32)

    def test_transpose(self) -> None:
        # Base case: no transposition
        args = self.get_default_tensor_args()
        x = torch.randn((1, 3, 32, 32))
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (1, 3, 32, 32)

        # Transpose (3 dimensions)
        args["transpose"] = True
        x = torch.randn((3, 32, 32))
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (32, 32, 3)

        # Transpose (4 dimensions)
        x = torch.randn((1, 3, 32, 32))
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.shape == (1, 32, 32, 3)

    def test_dtype(self) -> None:
        # Base case: no change.
        args = self.get_default_tensor_args()
        x = torch.randint(0, 10, (32, 32), dtype=torch.int32)
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.dtype == np.int32

        # Convert
        args["dtype"] = torch.int32
        x = torch.randn((32, 32), dtype=torch.float)
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.dtype == np.int32

    def test_as_uint8(self) -> None:
        # Base case: no change.
        args = self.get_default_tensor_args()
        x = torch.FloatTensor((0.0, 1.0))
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.dtype == np.float32
        assert np.max(ret) == 1.0
        assert np.min(ret) == 0.0

        # Change to uint8
        args["as_uint8"] = True
        ret = hldf.tensor_to_numpy(x, **args)
        assert isinstance(ret, np.ndarray)
        assert ret.dtype == np.uint8
        assert np.max(ret) == 255
        assert np.min(ret) == 0

    def test_defaults(self) -> None:
        x = torch.rand(1, 3, 32, 32)
        ret = hldf.tensor_to_numpy(x)
        assert x.shape == (1, 3, 32, 32)
        assert isinstance(ret, np.ndarray)
        assert ret.dtype == np.uint8
        assert ret.shape == (32, 32, 3)
        assert np.min(ret) == 0
        assert np.max(ret) <= 255


class TestDataModule:
    def get_expected_train_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def get_expected_valid_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[809, 833, 629], [694, 632, 678]]),
            torch.tensor([[359, 639, 636], [553, 266, 928]]),
            torch.tensor([[500, 324, 687], [963, 738, 996]]),
            torch.tensor([[421, 568, 705], [15, 841, 180]]),
            torch.tensor([[766, 97, 292], [23, 322, 577]]),
            torch.tensor([[269, 806, 472], [757, 651, 429]]),
            torch.tensor([[340, 810, 512], [440, 610, 126]]),
            torch.tensor([[92, 228, 424], [686, 769, 903]]),
        ]

    def get_expected_test_batches(self) -> list[torch.Tensor]:
        return [
            torch.tensor([[359, 639, 636], [963, 738, 996]]),
            torch.tensor([[766, 97, 292], [500, 324, 687]]),
            torch.tensor([[15, 841, 180], [553, 266, 928]]),
            torch.tensor([[269, 806, 472], [694, 632, 678]]),
            torch.tensor([[92, 228, 424], [421, 568, 705]]),
            torch.tensor([[686, 769, 903], [440, 610, 126]]),
            torch.tensor([[809, 833, 629], [340, 810, 512]]),
            torch.tensor([[23, 322, 577], [757, 651, 429]]),
        ]

    def check_batches(self, exp: list[torch.Tensor], val: list[torch.Tensor]) -> None:
        for b, exp_b in zip(exp, val, strict=True):
            assert torch.all(b == exp_b)

    def prepare(
        self, split: data.DatasetSplit, skip_seed: bool = False
    ) -> tuple[tud.DataLoader, tud.Sampler]:
        if not skip_seed:
            rng.seed_rngs()

        datamodule = SampleDataModule()
        datamodule.setup()

        if split == data.DatasetSplit.TRAIN:
            return core.get_from_optional(datamodule.train_dataloader())
        if split == data.DatasetSplit.VALID:
            return core.get_from_optional(datamodule.valid_dataloader())
        return core.get_from_optional(datamodule.test_dataloader())

    def test_worker_seeding(self) -> None:
        dataloader, _ = self.prepare(data.DatasetSplit.TRAIN)

        batches = []
        for batch in dataloader:
            batches.append(batch)

        exp_batches = self.get_expected_train_batches()
        self.check_batches(exp_batches, batches)

    def test_resume_random(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID)
        assert isinstance(sampler, hlds.ResumableRandomSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_valid_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.VALID, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableRandomSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_resume_sequential(self) -> None:
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)

        half_step = len(dataloader) // 2
        exp_batches = self.get_expected_test_batches()

        batches = []
        sampler.set_epoch(0)
        for step, batch in enumerate(dataloader):
            if step >= half_step:
                break
            batches.append(batch)

        self.check_batches(exp_batches[:half_step], batches)

        # Grab the state and clear everything.
        rng_state = rng.get_rng_state_dict()

        del dataloader, sampler
        batches = []
        rng.seed_rngs(0)

        # Restore everything.
        rng.load_rng_state_dict(rng_state)
        dataloader, sampler = self.prepare(data.DatasetSplit.TEST, skip_seed=True)
        assert isinstance(sampler, hlds.ResumableSequentialSampler)
        sampler.set_epoch(0)
        sampler.start_iter = half_step

        # Continue.
        for batch in dataloader:
            batches.append(batch)

        self.check_batches(exp_batches[half_step:], batches)

    def test_sampler_registry(self, check_registry) -> None:
        check_registry(
            hlds.SAMPLER_REGISTRY,
            [
                "ResumableRandomSampler",
                "ResumableSequentialSampler",
                "ResumableDistributedSampler",
            ],
        )

    def test_sampler_create(self, check_create_function) -> None:
        check_create_function(hlds.SAMPLER_REGISTRY, hlds.create_sampler)

    def test_dataset_registry(self, check_registry) -> None:
        check_registry(
            data.DATASET_REGISTRY,
            [
                "RandomDataset",
                "SequentialDataset",
            ],
        )

    def test_dataset_create(self, check_create_function) -> None:
        check_create_function(data.DATASET_REGISTRY, data.create_dataset)

    def test_collate_fn_registry(self, check_registry) -> None:
        check_registry(
            data.COLLATE_FN_REGISTRY,
            [
                "default_collate",
            ],
        )

    def test_collate_fn_create(self, check_create_function) -> None:
        data.COLLATE_FN_REGISTRY.register(dummy_collate_fn)
        ret = data.create_collate_fn("dummy_collate_fn")
        val = ret(1)
        assert val == 2

        # Ignore additional arguments when creating functions
        data.create_collate_fn("dummy_collate_fn", a=1)

        # Ensure arguments are forwarded when creating a type
        ret = data.create_collate_fn("SampleCollateFn", seed=1)
        assert ret() == 2

    def test_dataset_getters(self) -> None:
        datamodule = SampleDataModule()
        datamodule.setup()

        assert datamodule.train_dataset is not None
        assert datamodule.valid_dataset is not None
        assert datamodule.test_dataset is not None

        assert isinstance(datamodule.train_dataset, RandomDataset)
        assert isinstance(datamodule.valid_dataset, SequentialDataset)
        assert isinstance(datamodule.test_dataset, SequentialDataset)


if __name__ == "__main__":
    t = TestDataModule()
    t.test_resume_sequential()



import typing

from torch import nn

from helios import core


def _register_default_losses(registry: core.Registry) -> None:
    """
    Register the Torch default loss functions.

    List was obtained from https://pytorch.org/docs/stable/nn.html#loss-functions

    Args:
        registry: the loss registry.
    """
    registry.register(nn.L1Loss)
    registry.register(nn.MSELoss)
    registry.register(nn.CrossEntropyLoss)
    registry.register(nn.CTCLoss)
    registry.register(nn.NLLLoss)
    registry.register(nn.PoissonNLLLoss)
    registry.register(nn.GaussianNLLLoss)
    registry.register(nn.KLDivLoss)
    registry.register(nn.BCELoss)
    registry.register(nn.BCEWithLogitsLoss)
    registry.register(nn.MarginRankingLoss)
    registry.register(nn.HingeEmbeddingLoss)
    registry.register(nn.MultiLabelMarginLoss)
    registry.register(nn.HuberLoss)
    registry.register(nn.SmoothL1Loss)
    registry.register(nn.SoftMarginLoss)
    registry.register(nn.MultiLabelSoftMarginLoss)
    registry.register(nn.CosineEmbeddingLoss)
    registry.register(nn.MultiMarginLoss)
    registry.register(nn.TripletMarginLoss)
    registry.register(nn.TripletMarginWithDistanceLoss)


LOSS_REGISTRY = core.Registry("loss")
"""
Global instance of the registry for loss functions.

By default, the registry contains the following losses:

.. list-table:: Optimizers
    :header-rows: 1

    * - Loss
      - Name
    * - ``torch.nn.L1Loss``
      - L1Loss
    * - ``torch.nn.MSELoss``
      - MSELoss
    * - ``torch.nn.CrossEntropyLoss``
      - CrossEntropyLoss
    * - ``torch.nn.CTCLoss``
      - CTCLoss
    * - ``torch.nn.NLLLoss``
      - NLLLoss
    * - ``torch.nn.PoissonNLLLoss``
      - PoissonNLLLoss
    * - ``torch.nn.GaussianNLLLoss``
      - GaussianNLLLoss
    * - ``torch.nn.KLDivLoss``
      - KLDivLoss
    * - ``torch.nn.BCELoss``
      - BCELoss
    * - ``torch.nn.BCEWithLogitsLoss``
      - BCEWithLogitsLoss
    * - ``torch.nn.MarginRankingLoss``
      - MarginRankingLoss
    * - ``torch.nn.HingeEmbeddingLoss``
      - HingeEmbeddingLoss
    * - ``torch.nn.MultiLabelMarginLoss``
      - MultiLabelMarginLoss
    * - ``torch.nn.HuberLoss``
      - HuberLoss
    * - ``torch.nn.SmoothL1Loss``
      - SmoothL1Loss
    * - ``torch.nn.SoftMarginLoss``
      - SoftMarginLoss
    * - ``torch.nn.MultiLabelMarginLoss``
      - MultiLabelMarginLoss
    * - ``torch.nn.CosineEmbeddingLoss``
      - CosineEmbeddingLoss
    * - ``torch.nn.MultiMarginLoss``
      - MultiMarginLoss
    * - ``torch.nn.TripletMarginLoss``
      - TripletMarginLoss
    * - ``torch.nn.TripletMarginWithDistanceLoss``
      - TripletMarginWithDistanceLoss

Example:
    .. code-block:: python

        import helios.losses as hll

        # This automatically registers your loss function.
        @hll.LOSS_REGISTRY.register
        class MyLoss:
            ...

        # Alternatively you can manually register a loss function like this:
        hll.LOSS_REGISTRY.register(MyLoss)
"""
_register_default_losses(LOSS_REGISTRY)


def create_loss(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the loss function for the given type.

    Args:
        type_name: the type of the loss to create.
        args: positional arguments to pass into the loss.
        kwargs: keyword arguments to pass into the loss.
    Returns:
        The loss function.
    """
    return LOSS_REGISTRY.get(type_name)(*args, **kwargs)



import copy
import enum
import pathlib
import pickle
import typing

import numpy as np
import numpy.typing as npt
import pytest
import torch
from torch import nn
from torch.utils import data as tud

import helios.core as hlc
import helios.plugins as hlp
import helios.trainer as hlt
from helios import data
from helios import model as hlm
from helios.core import rng
from helios.data import functional as F

# Ignore the use of private members so we can test them correctly.
# ruff: noqa: SLF001

DATASET_SIZE = 10


class SimpleNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)

    def forward(self, x):
        return self.conv1(x)


class RandomDataset(tud.Dataset):
    def __getitem__(self, index):
        gen = rng.get_default_numpy_rng().generator
        return gen.integers(0, 100, 3)

    def __len__(self):
        return DATASET_SIZE


class RandomDatamodule(data.DataModule):
    def setup(self) -> None:
        params = data.DataLoaderParams(
            batch_size=1, num_workers=0, random_seed=rng.get_default_seed()
        )

        self._train_dataset = self._create_dataset(RandomDataset(), params)
        self._valid_dataset = self._create_dataset(RandomDataset(), params)
        self._test_dataset = self._create_dataset(RandomDataset(), params)


class CheckFunModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("test-model")
        self.called_train_funs: dict[str, bool] = {
            "setup": False,
            "train": False,
            "on_training_start": False,
            "on_training_epoch_start": False,
            "on_training_batch_start": False,
            "train_step": False,
            "on_training_batch_end": False,
            "on_training_epoch_end": False,
            "on_training_end": False,
            "eval": False,
            "on_validation_start": False,
            "on_validation_batch_start": False,
            "valid_step": False,
            "on_validation_batch_end": False,
            "on_validation_end": False,
            "have_metrics_improved": False,
            "should_training_stop": False,
        }

        self.called_test_funs: dict[str, bool] = {
            "load_for_testing": False,
            "eval": False,
            "on_testing_start": False,
            "on_testing_batch_start": False,
            "test_step": False,
            "on_testing_batch_end": False,
            "on_testing_end": False,
        }

    def setup(self, fast_init: bool = False) -> None:
        self.called_train_funs["setup"] = True

    def load_for_testing(self) -> None:
        self.called_test_funs["load_for_testing"] = True

    def train(self) -> None:
        assert self.called_train_funs["setup"]
        self.called_train_funs["train"] = True

    def on_training_start(self) -> None:
        self.called_train_funs["on_training_start"] = True

    def on_training_epoch_start(self, current_epoch) -> None:
        assert self.called_train_funs["train"]
        assert self.called_train_funs["on_training_start"]
        self.called_train_funs["on_training_epoch_start"] = True

    def on_training_batch_start(self, state) -> None:
        assert self.called_train_funs["on_training_epoch_start"]
        self.called_train_funs["on_training_batch_start"] = True

    def train_step(self, batch, state) -> None:
        assert self.called_train_funs["on_training_batch_start"]
        self.called_train_funs["train_step"] = True

    def on_training_batch_end(self, state, should_log: bool = False) -> None:
        assert self.called_train_funs["train_step"]
        self.called_train_funs["on_training_batch_end"] = True

    def on_training_epoch_end(self, current_epoch) -> None:
        assert self.called_train_funs["on_training_epoch_start"]
        self.called_train_funs["on_training_epoch_end"] = True

    def on_training_end(self) -> None:
        assert self.called_train_funs["on_training_start"]
        self.called_train_funs["on_training_end"] = True

    def eval(self) -> None:
        self.called_train_funs["eval"] = True
        self.called_test_funs["eval"] = True

    def on_validation_start(self, validation_cycle) -> None:
        assert self.called_train_funs["eval"]
        self.called_train_funs["on_validation_start"] = True

    def on_validation_batch_start(self, step) -> None:
        assert self.called_train_funs["eval"]
        assert self.called_train_funs["on_validation_start"]
        self.called_train_funs["on_validation_batch_start"] = True

    def valid_step(self, batch, step) -> None:
        assert self.called_train_funs["on_validation_batch_start"]
        self.called_train_funs["valid_step"] = True

    def on_validation_batch_end(self, step) -> None:
        assert self.called_train_funs["valid_step"]
        self.called_train_funs["on_validation_batch_end"] = True

    def on_validation_end(self, cycle) -> None:
        assert self.called_train_funs["on_validation_start"]
        self.called_train_funs["on_validation_end"] = True

    def have_metrics_improved(self) -> bool:
        assert self.called_train_funs["on_validation_end"]
        self.called_train_funs["have_metrics_improved"] = True
        return True

    def should_training_stop(self) -> bool:
        self.called_train_funs["should_training_stop"] = True
        return False

    def on_testing_start(self) -> None:
        self.called_test_funs["on_testing_start"] = True

    def on_testing_batch_start(self, step) -> None:
        assert self.called_test_funs["eval"]
        assert self.called_test_funs["on_testing_start"]
        self.called_test_funs["on_testing_batch_start"] = True

    def test_step(self, batch, step) -> None:
        assert self.called_test_funs["on_testing_batch_start"]
        self.called_test_funs["test_step"] = True

    def on_testing_batch_end(self, step) -> None:
        assert self.called_test_funs["test_step"]
        self.called_test_funs["on_testing_batch_end"] = True

    def on_testing_end(self) -> None:
        assert self.called_test_funs["on_testing_batch_end"]
        self.called_test_funs["on_testing_end"] = True


class RestartModel(hlm.Model):
    def __init__(self, val_count: int = -1) -> None:
        super().__init__("test-restart")
        self.batches: list[npt.NDArray] = []
        self.val_count = val_count

    def setup(self, fast_init: bool = False) -> None:
        pass

    def on_training_batch_start(self, state: hlt.TrainingState) -> None:
        if self.val_count == state.validation_cycles:
            raise RuntimeError("stop")

    def train_step(self, batch: torch.Tensor, state) -> None:
        as_np = F.tensor_to_numpy(batch)
        self.batches.append(as_np)


class Values(enum.Enum):
    A = enum.auto()
    B = enum.auto()


class CheckpointModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("test-checkpoint")

        self._state = {
            "a": 1,
            "b": 2.0,
            "c": "x",
            "value": Values.A,
        }

    def setup(self, fast_init: bool = False) -> None:
        pass

    def types_for_safe_load(self) -> list[typing.Callable | tuple[typing.Callable, str]]:
        return [Values]

    def state_dict(self) -> dict[str, typing.Any]:
        return self._state

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        assert self._state == state_dict


class AccumulationModel(hlm.Model):
    def __init__(self, accumulation_steps: int = 1) -> None:
        super().__init__("test-accumulation")
        self.accumulation_steps = accumulation_steps
        self.global_steps: int = 0
        self.accumulated_steps: int = 0

    def setup(self, fast_init: bool = False) -> None:
        pass

    def train_step(self, batch: torch.Tensor, state: hlt.TrainingState) -> None:
        self.global_steps += 1
        if self.global_steps % self.accumulation_steps == 0:
            self.accumulated_steps += 1

        assert state.global_iteration == self.global_steps
        assert state.current_iteration == self.accumulated_steps


class ExceptionModel(hlm.Model):
    def __init__(self, exc_type: type[Exception]) -> None:
        super().__init__("test-exception")
        self._exc_type = exc_type

    def setup(self, fast_init: bool = False) -> None:
        pass

    def on_training_start(self) -> None:
        raise self._exc_type("error")

    def on_testing_start(self) -> None:
        raise self._exc_type("error")


class EmptyModel(hlm.Model):
    def __init__(self) -> None:
        super().__init__("empty")

    def setup(self, fast_init: bool = False) -> None:
        pass


class EmptyPlugin(hlp.Plugin):
    def __init__(self):
        super().__init__("empty")

    def setup(self):
        pass


class CheckpointPlugin(hlp.Plugin):
    def __init__(self):
        super().__init__("chkpt")
        self._state = {
            "a": 1,
            "b": 2.0,
            "c": "x",
        }

    def setup(self):
        pass

    def configure_trainer(self, trainer: hlt.Trainer) -> None:
        self._register_in_trainer(trainer)

    def state_dict(self) -> dict[str, typing.Any]:
        return self._state

    def load_state_dict(
        self, state_dict: dict[str, typing.Any], fast_init: bool = False
    ) -> None:
        assert self._state == state_dict


class OverrideFlagsPlugin(hlp.Plugin):
    def __init__(
        self,
        training_batch: bool = False,
        validation_batch: bool = False,
        testing_batch: bool = False,
        should_training_stop: bool = False,
    ):
        super().__init__("override")
        self._overrides.training_batch = training_batch
        self._overrides.validation_batch = validation_batch
        self._overrides.testing_batch = testing_batch
        self._overrides.should_training_stop = should_training_stop

    def setup(self):
        pass


class CheckFunPlugin(hlp.Plugin):
    name: str = "check"

    def __init__(self):
        super().__init__(self.name)
        self._overrides.training_batch = True
        self._overrides.validation_batch = True
        self._overrides.testing_batch = True
        self._overrides.should_training_stop = True

        self.called_train_funs: dict[str, bool] = {
            "setup": False,
            "on_training_start": False,
            "process_training_batch": False,
            "on_training_end": False,
            "on_validation_start": False,
            "process_validation_batch": False,
            "on_validation_end": False,
            "should_training_stop": False,
        }

        self.called_test_funs: dict[str, bool] = {
            "on_testing_start": False,
            "process_testing_batch": False,
            "on_testing_end": False,
        }

    def setup(self) -> None:
        self.called_train_funs["setup"] = True

    def on_training_start(self) -> None:
        self.called_train_funs["on_training_start"] = True

    def process_training_batch(self, batch, state) -> typing.Any:
        self.called_train_funs["process_training_batch"] = True
        return batch

    def on_training_end(self) -> None:
        self.called_train_funs["on_training_end"] = True

    def on_validation_start(self, validation_cycle) -> None:
        self.called_train_funs["on_validation_start"] = True

    def process_validation_batch(self, batch, step) -> typing.Any:
        self.called_train_funs["process_validation_batch"] = True
        return batch

    def on_validation_end(self, validation_cycle) -> None:
        self.called_train_funs["on_validation_end"] = True

    def should_training_stop(self) -> bool:
        self.called_train_funs["should_training_stop"] = True
        return False

    def on_testing_start(self) -> None:
        self.called_test_funs["on_testing_start"] = True

    def process_testing_batch(self, batch, step) -> typing.Any:
        self.called_test_funs["process_testing_batch"] = True
        return batch

    def on_testing_end(self) -> None:
        self.called_test_funs["on_testing_end"] = True


class CheckPluginModel(hlm.Model):
    def __init__(self):
        super().__init__("test-plugin")

    def _get_plugin(self) -> CheckFunPlugin:
        plugin = self.trainer.plugins[CheckFunPlugin.name]
        assert isinstance(plugin, CheckFunPlugin)
        return plugin

    def setup(self, fast_init: bool = False) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["setup"]

    def on_training_start(self) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["on_training_start"]

    def train_step(self, batch, state) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["process_training_batch"]

    def on_training_end(self) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["on_training_end"]

    def on_validation_start(self, cycle) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["on_validation_start"]

    def valid_step(self, batch, step) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["process_validation_batch"]

    def on_validation_end(self, cycle) -> None:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["on_validation_end"]

    def should_training_stop(self) -> bool:
        plugin = self._get_plugin()
        assert plugin.called_train_funs["should_training_stop"]
        return False

    def on_testing_start(self) -> None:
        plugin = self._get_plugin()
        assert plugin.called_test_funs["on_testing_start"]

    def test_step(self, batch, step) -> None:
        plugin = self._get_plugin()
        assert plugin.called_test_funs["process_testing_batch"]

    def on_testing_end(self) -> None:
        plugin = self._get_plugin()
        assert plugin.called_test_funs["on_testing_end"]


class TestTrainingUnit:
    def test_from_str(self) -> None:
        assert hlt.TrainingUnit.from_str("epoch") == hlt.TrainingUnit.EPOCH
        assert hlt.TrainingUnit.from_str("iteration") == hlt.TrainingUnit.ITERATION

        with pytest.raises(ValueError):
            hlt.TrainingUnit.from_str("foo")


class TestTrainer:
    def test_register_types(self, tmp_path: pathlib.Path) -> None:
        test_dict = {"state": hlt.TrainingState(), "path": tmp_path}
        out_path = tmp_path / "register.pth"
        torch.save(test_dict, out_path)

        torch.serialization.clear_safe_globals()
        with pytest.raises(pickle.UnpicklingError):
            hlc.safe_torch_load(out_path)

        hlt.register_trainer_types_for_safe_load()
        reg_types = torch.serialization.get_safe_globals()
        exp_types = hlt.get_trainer_safe_types_for_load()
        assert set(reg_types) == set(exp_types)

        ret_dict = hlc.safe_torch_load(out_path)
        assert test_dict == ret_dict

    def test_find_chkpt(self, tmp_path: pathlib.Path) -> None:
        test_cases = [
            (tmp_path / "test_epoch_1_iter_10.pth", True),
            (tmp_path / "test.txt", False),
            (tmp_path / "test.pth", False),
            (tmp_path / "test_epoch_1.pth", False),
            (tmp_path / "test_iter_1.pth", False),
        ]

        for case in test_cases:
            file, exp = case
            file.touch()
            ret = hlt.find_last_checkpoint(tmp_path)
            if exp:
                assert ret is not None
            else:
                assert ret is None
            file.unlink()

    def check_trainer_flags(self, **kwargs) -> None:
        with pytest.raises(ValueError):
            hlt.Trainer(**kwargs)

    def test_trainer_validate(self, tmp_path: pathlib.Path) -> None:
        hlt.Trainer()
        invalid_root = tmp_path / "test.txt"
        invalid_root.touch()

        self.check_trainer_flags(total_steps=3.0)
        self.check_trainer_flags(chkpt_frequency=0)
        self.check_trainer_flags(print_frequency=0)
        self.check_trainer_flags(valid_frequency=0)
        self.check_trainer_flags(enable_deterministic=True, enable_cudnn_benchmark=True)
        self.check_trainer_flags(total_steps=float("inf"), early_stop_cycles=0)
        self.check_trainer_flags(enable_tensorboard=True, run_path=None)
        self.check_trainer_flags(enable_tensorboard=True, run_path=invalid_root)
        self.check_trainer_flags(enable_file_logging=True, log_path=None)
        self.check_trainer_flags(enable_file_logging=True, log_path=invalid_root)
        self.check_trainer_flags(use_cpu=True, gpus=[1])

    def test_trainer_device_flags(self) -> None:
        t = hlt.Trainer(use_cpu=True)
        assert t._use_cpu
        assert t._device == torch.device("cpu")
        assert t._map_loc == {"cuda:0": "cpu"}
        assert t._gpu_ids == []
        assert not t._is_distributed

        if torch.cuda.is_available():
            t = hlt.Trainer()
            devices = list(range(torch.cuda.device_count()))
            assert not t._use_cpu
            assert t._gpu_ids == devices
            assert t._is_distributed if len(devices) > 1 else not t._is_distributed

            with pytest.raises(ValueError):
                devs = copy.deepcopy(devices)
                devs.append(10)
                hlt.Trainer(gpus=devs)

            with pytest.raises(ValueError):
                hlt.Trainer(gpus=[len(devices)])

    def check_map_location(
        self,
        out_path: pathlib.Path,
        start_device: torch.device,
        train_args: dict[str, typing.Any],
    ) -> None:
        net = SimpleNet().to(start_device)
        torch.save(net.state_dict(), out_path)

        t = hlt.Trainer(**train_args)
        t.model = EmptyModel()
        t._configure_env()
        net = SimpleNet().to(t._device)
        data = hlc.safe_torch_load(out_path, map_location=t._map_loc)
        net.load_state_dict(data)

    def test_map_location(self, tmp_path: pathlib.Path) -> None:
        self.check_map_location(
            tmp_path / "cpu_to_cpu.pth", torch.device("cpu"), {"use_cpu": True}
        )

        if torch.cuda.is_available():
            self.check_map_location(
                tmp_path / "cpu_to_cuda.pth", torch.device("cpu"), {"gpus": [0]}
            )
            self.check_map_location(
                tmp_path / "cuda_to_cpu.pth", torch.device("cuda:0"), {"use_cpu": True}
            )

            available_ids = list(range(torch.cuda.device_count()))
            if len(available_ids) > 1:
                self.check_map_location(
                    tmp_path / "cpu_to_cuda1.pth", torch.device("cpu"), {"gpus": [1]}
                )
                self.check_map_location(
                    tmp_path / "cuda1_to_cpu.pth",
                    torch.device("cuda:1"),
                    {"use_cpu": True},
                )
                self.check_map_location(
                    tmp_path / "cuda0_to_cuda1.pth", torch.device("cuda:0"), {"gpus": [1]}
                )
                self.check_map_location(
                    tmp_path / "cuda1_to_cuda0.pth", torch.device("cuda:1"), {"gpus": [0]}
                )

    def check_training_loops(
        self,
        trainer: hlt.Trainer,
        num_chkpts: int = 0,
        chkpt_root: pathlib.Path | None = None,
        fit: bool = True,
    ) -> None:
        datamodule = RandomDatamodule()
        model = CheckFunModel()

        called_funs: dict[str, bool]
        if fit:
            assert trainer.fit(model, datamodule)
            called_funs = model.called_train_funs
        else:
            assert trainer.test(model, datamodule)
            called_funs = model.called_test_funs

        for _, seen in called_funs.items():
            assert seen

        if fit:
            assert chkpt_root is not None
            assert num_chkpts == len(list(chkpt_root.glob("*.pth")))

            for chkpt_path in chkpt_root.glob("*.pth"):
                state_dict = hlc.safe_torch_load(chkpt_path)
                assert trainer._validate_state_dict(state_dict)

                # Add an extra key
                state_dict["tmp"] = "foo"
                assert trainer._validate_state_dict(state_dict)
                state_dict.pop("tmp")

                # Remove one of the valid keys
                state_dict.pop("rng")
                assert not trainer._validate_state_dict(state_dict)

    def test_fit_iter(self, tmp_path: pathlib.Path) -> None:
        self.check_training_loops(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.ITERATION,
                total_steps=10,
                valid_frequency=5,
                chkpt_frequency=5,
                use_cpu=True,
                enable_progress_bar=False,
                chkpt_root=tmp_path / "chkpt",
            ),
            2,
            tmp_path / "chkpt/test-model",
        )

    def test_fit_epoch(self, tmp_path: pathlib.Path) -> None:
        self.check_training_loops(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.EPOCH,
                total_steps=2,
                valid_frequency=1,
                chkpt_frequency=1,
                use_cpu=True,
                chkpt_root=tmp_path / "chkpt",
            ),
            2,
            tmp_path / "chkpt/test-model",
        )

    def test_testing(self) -> None:
        self.check_training_loops(hlt.Trainer(use_cpu=True), fit=False)

    def get_restart_trainer(
        self, unit: hlt.TrainingUnit, chkpt_root: pathlib.Path
    ) -> hlt.Trainer:
        if unit == hlt.TrainingUnit.ITERATION:
            total_steps = DATASET_SIZE
            valid_frequency = DATASET_SIZE // 2
        else:
            total_steps = 2
            valid_frequency = 1

        return hlt.Trainer(
            train_unit=unit,
            total_steps=total_steps,
            valid_frequency=valid_frequency,
            chkpt_frequency=valid_frequency,
            use_cpu=True,
            chkpt_root=chkpt_root,
        )

    def get_restart_model_and_datamodule(
        self, val_count: int = -1
    ) -> tuple[RandomDatamodule, RestartModel]:
        return RandomDatamodule(), RestartModel(val_count)

    def clear_chkpts(self, chkpt_root: pathlib.Path) -> None:
        for chkpt in chkpt_root.glob("*.pth"):
            chkpt.unlink()

    def check_batches(
        self, exp_batches: list[npt.NDArray], ret_batches: list[npt.NDArray]
    ) -> None:
        assert len(exp_batches) == len(ret_batches)
        for exp, ret in zip(exp_batches, ret_batches, strict=True):
            assert np.all(exp == ret)

    def check_restart_trainer(
        self, unit: hlt.TrainingUnit, tmp_path: pathlib.Path
    ) -> None:
        datamodule, model = self.get_restart_model_and_datamodule()
        trainer = self.get_restart_trainer(unit, tmp_path)
        chkpt_root = tmp_path / model.save_name

        assert trainer.fit(model, datamodule)
        batches = model.batches

        self.clear_chkpts(chkpt_root)
        del datamodule, model, trainer

        datamodule, model = self.get_restart_model_and_datamodule(1)
        trainer = self.get_restart_trainer(unit, tmp_path)
        assert not trainer.fit(model, datamodule)

        # Clear out everything again and restart.
        ret_batches = model.batches
        del datamodule, trainer, model

        datamodule, model = self.get_restart_model_and_datamodule()
        trainer = self.get_restart_trainer(unit, tmp_path)
        model.batches = ret_batches
        assert trainer.fit(model, datamodule)

        self.check_batches(batches, model.batches)

    def test_restart_iter(self, tmp_path: pathlib.Path) -> None:
        self.check_restart_trainer(hlt.TrainingUnit.ITERATION, tmp_path)

    def test_restart_epoch(self, tmp_path: pathlib.Path) -> None:
        self.check_restart_trainer(hlt.TrainingUnit.EPOCH, tmp_path)

    def check_accumulation(self, num_steps: int) -> None:
        datamodule = RandomDatamodule()
        model = AccumulationModel(num_steps)
        trainer = hlt.Trainer(
            train_unit=hlt.TrainingUnit.ITERATION,
            total_steps=20,
            use_cpu=True,
            accumulation_steps=num_steps,
        )

        assert trainer.fit(model, datamodule)

    def test_accumulation(self) -> None:
        self.check_accumulation(1)
        self.check_accumulation(2)
        self.check_accumulation(4)
        self.check_accumulation(5)
        self.check_accumulation(10)

    def check_exception(
        self,
        exc_type: type[Exception],
        trainer: hlt.Trainer,
        fit: bool,
        raised_as_runtime: bool = False,
    ) -> None:
        datamodule = RandomDatamodule()
        model = ExceptionModel(exc_type)

        if not raised_as_runtime:
            with pytest.raises(exc_type):
                if fit:
                    trainer.fit(model, datamodule)
                else:
                    trainer.test(model, datamodule)
        else:
            if fit:
                assert not trainer.fit(model, datamodule)
            else:
                assert not trainer.test(model, datamodule)

    def test_trainer_exceptions(self) -> None:
        exception_types = [ValueError, RuntimeError, KeyError]
        trainer = hlt.Trainer()
        trainer.train_exceptions = exception_types[:2]
        for exc_type in trainer.train_exceptions:
            self.check_exception(exc_type, trainer, fit=True)
        self.check_exception(
            exception_types[-1], trainer, fit=True, raised_as_runtime=True
        )

        trainer.train_exceptions = []
        trainer.test_exceptions = exception_types[:2]
        for exc_type in trainer.test_exceptions:
            self.check_exception(exc_type, trainer, fit=False)
        self.check_exception(
            exception_types[-1], trainer, fit=False, raised_as_runtime=True
        )

    def test_append_plugins(self) -> None:
        trainer = hlt.Trainer()

        trainer.plugins["empty"] = EmptyPlugin()
        trainer._validate_plugins()

        batch_flags = {
            "training_batch": False,
            "validation_batch": False,
            "testing_batch": False,
            "should_training_stop": False,
        }

        for flag_name in batch_flags:
            batch_flags[flag_name] = True
            trainer.plugins[f"override_base_{flag_name}"] = OverrideFlagsPlugin(
                **batch_flags
            )

            with pytest.raises(ValueError):
                trainer.plugins[f"override_dup_{flag_name}"] = OverrideFlagsPlugin(
                    **batch_flags
                )
                trainer._validate_plugins()

            batch_flags[flag_name] = False

    def check_plugin_functions(self, trainer: hlt.Trainer, fit: bool = True) -> None:
        datamodule = RandomDatamodule()
        model = CheckPluginModel()

        name = CheckFunPlugin.name
        trainer.plugins[name] = CheckFunPlugin()

        called_funs: dict[str, bool]
        if fit:
            assert trainer.fit(model, datamodule)
            called_funs = trainer.plugins[name].called_train_funs  # type: ignore[attr-defined]
        else:
            assert trainer.test(model, datamodule)
            called_funs = trainer.plugins[name].called_test_funs  # type: ignore[attr-defined]

        for _, seen in called_funs.items():
            assert seen

    def test_plugin_functions(self) -> None:
        self.check_plugin_functions(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.ITERATION,
                total_steps=10,
                valid_frequency=10,
                use_cpu=True,
            )
        )

        self.check_plugin_functions(
            hlt.Trainer(
                train_unit=hlt.TrainingUnit.EPOCH,
                total_steps=1,
                valid_frequency=1,
                use_cpu=True,
            )
        )

        self.check_plugin_functions(hlt.Trainer(use_cpu=True), fit=False)

    def test_checkpoints(self, tmp_path: pathlib.Path) -> None:
        datamodule = RandomDatamodule()
        model = CheckpointModel()
        plugin = CheckpointPlugin()

        trainer = hlt.Trainer(
            train_unit=hlt.TrainingUnit.ITERATION,
            total_steps=10,
            valid_frequency=10,
            chkpt_root=tmp_path,
            chkpt_frequency=10,
            use_cpu=True,
        )
        plugin.configure_trainer(trainer)
        assert trainer.fit(model, datamodule)
        safe_globals = torch.serialization.get_safe_globals()
        for t in model.types_for_safe_load():
            assert t in safe_globals

        chkpt_root = tmp_path / model._save_name
        chkpts = list(chkpt_root.glob("*.pth"))
        assert len(chkpts) == 1

        chkpt = hlc.safe_torch_load(chkpts[0])
        model.load_state_dict(chkpt["model"])
        plugin.load_state_dict(chkpt[plugin._plug_id])
